class FakeGPSModule {
  htmlelement = document.querySelector("#fakegps-addon");

  current_position = null;
  current_bearing = null;
  current_accuracy = null;
  current_speed = null;

  // Target Position
  target_position = null;

  // Default speeds in km/h
  default_speeds = {
    foot: 7,
    bike: 20,
    car: 50,
  };

  // Event handlers
  rightclick_handler = this.rightclick.bind(this);
  gps_updater = null; // Interval for updating the GPS position in local storage

  user_position_updater = (e) => {
    let center = e.target.getCenter();
    let next_position = turf.point([center.lng, center.lat]);
    let bearing = turf.bearing(this.current_position, next_position);
    if (turf.booleanEqual(this.current_position, this.target_position)) {
      map.map.off("render", this.user_position_updater);
      this.target_position = null;
      this.current_speed = 0;
      this.updateCurrentPosition(
        next_position,
        this.current_bearing,
        this.current_accuracy
      );
    } else {
      this.updateCurrentPosition(next_position, bearing, this.current_accuracy);
    }
  };

  constructor() {
    document.querySelector("#nav-fakegps").onclick = (e) => {
      e.preventDefault();
      map.switchModule("fakegps");
    };

    this.htmlelement
      .querySelectorAll("input[name=profile]")
      .forEach((element) => {
        element.onchange = (e) => {
          let selected_profile = e.target.value;
          this.htmlelement.querySelector("#speed").value =
            this.default_speeds[selected_profile];
        };
      });

    // Prevent non numbers in speed input field
    this.htmlelement
      .querySelector("#speed")
      .addEventListener("keypress", (e) => {
        if (e.which < 40 || e.which > 57) e.preventDefault();
      });

    this.htmlelement
      .querySelector("#accuracy")
      .addEventListener("change", (e) => {
        let new_accuracy = parseInt(
          this.htmlelement.querySelector("#accuracy").value
        );
        this.updateCurrentPosition(
          this.current_position,
          this.current_bearing,
          new_accuracy
        );
      });

    // Exit button
    this.htmlelement.querySelector(".exit > button").onclick = (e) => {
      map.switchModule("search");
    };
  }

  updateCurrentPosition(
    point = this.current_position,
    bearing = this.current_bearing,
    accuracy = this.current_accuracy
  ) {
    // Calc distance between points
    this.current_position = point;
    this.current_bearing = bearing;
    this.current_accuracy = accuracy;
    if (map.user_marker) {
      map.updateUserPosition(point, bearing, accuracy);
    } else {
      map.showUserPosition(point, bearing, accuracy);
    }
    if (this.htmlelement.querySelector("#enable-updates").checked)
      this.storeLocationUpdate(point, bearing, accuracy);
  }

  enable() {
    this.htmlelement.classList.add("active");
    document.getElementById("nav-opener").checked = false;
    this.updateState();

    let selected_profile = this.htmlelement.querySelector(
      "input[name=profile]:checked"
    ).value;
    this.htmlelement.querySelector("#speed").value =
      this.default_speeds[selected_profile];

    if (!this.current_accuracy) {
      this.current_accuracy = parseInt(
        this.htmlelement.querySelector("#accuracy").value
      );
    }

    if (this.current_position == null) {
      let point = map.map.getCenter();
      point = turf.point([point.lng, point.lat]);
      this.updateCurrentPosition(
        point,
        this.current_bearing,
        this.current_accuracy
      );
    }

    map.map.dragPan.disable();
    map.map.scrollZoom.disable();
    map.map.keyboard.disable();
    map.map.dragRotate.disable();
    // Update state when moving
    map.map.on("moveend", this.updateState);
    // Add Listener to right Click
    map.map.on("contextmenu", this.rightclick_handler);

    // start interval for updating the current position
    this.gps_updater = setInterval(() => {
      if (!this.target_position) {
        this.current_speed = 0;
        this.updateCurrentPosition();
      }
    }, 2500);
  }

  rightclick(event) {
    console.log(event);
    this.animatePosition(turf.point([event.lngLat.lng, event.lngLat.lat]));
  }

  animatePosition(target) {
    if (!this.current_position) {
      this.updateCurrentPosition(target, null, this.current_accuracy);
      return;
    }

    this.current_speed = parseInt(
      this.htmlelement.querySelector("#speed").value
    );

    let distance = turf.distance(this.current_position, target);

    if (distance > 2) {
      // Do not animate too long jumps
      map.map.jumpTo({ center: target.geometry.coordinates });
      this.updateCurrentPosition(target, null, this.current_accuracy);
      return;
    }

    let duration_ms = (distance / this.current_speed) * 60 * 60 * 1000;

    this.target_position = target;

    map.map.off("render", this.user_position_updater);
    map.map.on("render", this.user_position_updater);
    map.map.easeTo({
      center: target.geometry.coordinates,
      easing: (time) => time,
      duration: duration_ms,
    });
  }

  storeLocationUpdate(point, heading = null, accuracy) {
    let speed = this.current_speed;
    if (speed == 0) {
      heading = "NaN";
    } else {
      // Convert from km/h to m/s
      speed = speed * (5 / 18);
    }
    if (heading && heading != "NaN") {
      // The supplied heading is -180 to 180 degrees.
      // We need to convert it to 0 to 360 degrees
      if (heading < 0) {
        heading = 180 + (180 - Math.abs(heading));
      }
    }

    // Generate a random point according to the accuracy
    let circle = turf.circle(point, accuracy / 1000);
    let bbox = turf.bbox(circle);
    do {
      let random_x = (Math.random() * (bbox[2] - bbox[0])) + bbox[0];
      let random_y = (Math.random() * (bbox[3] - bbox[1])) + bbox[1];
      point = turf.point([random_x, random_y]);
    } while (!turf.booleanPointInPolygon(point, circle));

    // Fake Geolocation Coordinates as geolocation api would return them
    // According to https://developer.mozilla.org/en-US/docs/Web/API/GeolocationCoordinates
    let geolocation_data = {
      timestamp: Date.now(),
      coords: {
        longitude: point.geometry.coordinates[0],
        latitude: point.geometry.coordinates[1],
        altitude: null,
        accuracy: accuracy,
        heading: heading,
        speed: speed,
      },
    };

    localStorage.setItem("location", JSON.stringify(geolocation_data));
  }

  exit() {
    this.htmlelement.classList.remove("active");
    map.hideUserPosition();
    map.map.off("contextmenu", this.rightclick_handler);
    map.map.dragPan.enable();
    map.map.scrollZoom.enable();
    map.map.keyboard.enable();
    map.map.dragRotate.enable();
    if (this.gps_updater) clearInterval(this.gps_updater);
  }

  updateState() {
    let current_bounds = map.map.getBounds();
    let url = `/fakegps/${JSON.stringify(current_bounds.toArray())}`;
    let state = {
      module: "fakegps",
      mapposition: current_bounds,
      timestamp: Date.now(),
    };

    // Skip update if too frequest
    let udpate_rate_ms = 2500;
    if (
      window.history.state &&
      state.timestamp - window.history.state.timestamp < udpate_rate_ms
    ) {
      return;
    }

    if (window.location.pathname.match(/^\/fakegps/)) {
      window.history.replaceState(state, "", url);
    } else {
      window.history.pushState(state, "", url);
    }
  }
}
