class Result {
  data = null;
  error_message = null;
  error_code = null;

  static get ERROR_GEOLOCATION_PERMISSION_DENIED() {
    return 1;
  }
  static get ERROR_GEOLOCATION_INTERNAL_ERROR() {
    return 2;
  }
  static get ERROR_GEOLOCATION_TIMEOUT() {
    return 3;
  }

  source = "search";
  rank = 0;
  route = null;

  dom_element = null;
  marker = null;

  actions_enabled = false;

  geolocation_watch_id = null;
  navigation_available = false;

  constructor(
    result,
    enable_actions = false,
    source = "search",
    rank = 0,
    route = null,
    callback = null
  ) {
    this.actions_enabled = enable_actions;
    this.source = source;
    this.rank = rank;
    this.route = route;
    if (typeof result == "object") {
      if (result) this.data = result;
    } else if (typeof result == "string") {
      let detail_match = result.match(/^(N|W|R)(\d+)$/);
      if (detail_match) {
        // Add waypoint from lookup
        this.#lookupWaypoint(detail_match[1] + detail_match[2]).then(() => {
          this.#createDomElement();
          this.#createMarker();
          if (callback) callback();
          this.dom_element.dispatchEvent(new Event("load"));
          return this;
        });
      } else if (result.startsWith("gps")) {
        this.#getFromGPS(() => {
          this.#createDomElement();
          this.#createMarker();
          if (callback) callback();
          this.dom_element.dispatchEvent(new Event("load"));
        });
      }
    } else {
      throw new Error(`"${typeof result}" cannot be parsed to a result`);
    }
    this.#createDomElement();
    this.#createMarker();
    if (this.data && callback) callback();
  }

  showMarker() {
    if (this.marker && !this.error_message) {
      if (this.route && this.route.data) {
        let snapped_coordinates = this.route.data.paths[this.route.selected].snapped_waypoints.coordinates[0];
        this.marker.setLngLat(snapped_coordinates).addTo(map.map);
      } else {
        this.marker = this.marker
          .setLngLat([this.data.lon, this.data.lat])
          .addTo(map.map);
      }
    }
    if (this.source == "gps" && this.data && this.data.location) {
      map.showUserPosition(turf.point([this.data.location.coords.longitude, this.data.location.coords.latitude]), this.data.location.coords.heading, this.data.location.coords.accuracy);
    } else if (this.source == "gps") {
      map.hideUserPosition();
    }
  }

  removeMarker() {
    if (this.marker) this.marker = this.marker.remove();
    if (this.source == "gps" && this.data && this.data.location) map.hideUserPosition();
  }

  showGeoJson(sourcename_line, sourcename_fill) {
    if (!this.data || !this.data.geojson) return;
    if (map.map.getSource(sourcename_line) == undefined) {
      throw new Error(`Source "${sourcename_line}" does not exist on the map.`);
    }
    if (map.map.getSource(sourcename_fill) == undefined) {
      throw new Error(`Source "${sourcename_fill}" does not exist on the map.`);
    }
    let lineFeature = null;
    let fillFeature = null;

    // Create a combined boundingbox
    switch (this.data.geojson.type) {
      case "Polygon":
      case "MultiPolygon":
        fillFeature = {
          type: "Feature",
          properties: {},
          geometry: this.data.geojson,
        };
        break;
      case "LineString":
      case "MultiLineString":
        lineFeature = {
          type: "Feature",
          properties: {},
          geometry: this.data.geojson,
        };
        break;
    }
    if (lineFeature) {
      let line_data = map.map.getSource(sourcename_line)._data;
      line_data.features.push(lineFeature);
      map.map.getSource(sourcename_line).setData(line_data);
    }
    if (fillFeature) {
      let fill_data = map.map.getSource(sourcename_fill)._data;
      fill_data.features.push(fillFeature);
      map.map.getSource(sourcename_fill).setData(fill_data);
    }
  }

  toString() {
    switch (this.source) {
      case "gps":
        return `gps:${this.data.location.coords.longitude},${this.data.location.coords.latitude}`;
      default:
        return this.data
          ? this.data.osm_type[0].toUpperCase() + this.data.osm_id
          : null;
    }
  }

  updateRank(rank) {
    this.rank = rank;
    this.#createDomElement();
    this.#createMarker();
  }

  async #getFromGPS(callback) {
    let maximumAge = 500;
    let lastUpdate = Date.now() - maximumAge;
    let min_accuracy = 50; // Min 50m for Navigation
    let min_frequency = 2 * maximumAge; // Minimum Update frequency for navigation

    let frequency_high_enough = false;
    let accuracy_high_enough = false;
    let navigation_reported = false;

    let position_callback = (position) => {
      let coords = position.coords;
      let circle = turf.circle(
        [coords.longitude, coords.latitude],
        coords.accuracy / 1000
      );
      let result = {
        display_name: "Eigene Position",
        location: {
          timestamp: position.timestamp,
          coords: {
            latitude: coords.latitude,
            longitude: coords.longitude,
            altitude: coords.altitude,
            accuracy: coords.accuracy,
            altitudeAccuracy: coords.altitudeAccuracy,
            heading: coords.heading,
            speed: coords.speed
          }
        },
        lon: position.coords.longitude,
        lat: position.coords.latitude
      };

      let bbox = turf.bbox(circle);
      result.boundingbox = bbox;

      if (this.data == null) {
        this.data = result;
        if (callback) callback();
      } else {
        this.data = result;
        if (coords.accuracy <= min_accuracy) {
          accuracy_high_enough = true;
        }
        if (position.timestamp - lastUpdate < min_frequency) {
          frequency_high_enough = true;
        }

        if (!this.navigation_available && accuracy_high_enough && frequency_high_enough) {
          this.navigation_available = true;
          this.dom_element.dispatchEvent(new Event("navigation_available"));
        }
        if (position.timestamp - lastUpdate >= maximumAge) {
          lastUpdate = position.timestamp;
          this.dom_element.dispatchEvent(new CustomEvent("update", { detail: position }));
        }
      }
    };

    let error_callback = (error) => {
      console.log(error);
      map.gpsManager.clearWatch(this.geolocation_watch_id);
      this.geolocation_watch_id = null;
      if (this.data == null) {
        switch (error.code) {
          case 1:
            this.error_message = "Zugriff auf Standort abgelehnt";
            break;
          case 2:
            this.error_message = "Interner Fehler beim ermitteln der Position";
            break;
          case 3:
            this.error_message =
              "Standort konnte nicht schnell genug ermittelt werden";
            break;
        }
        this.error_code = error.code;
        callback();
      } else {
        this.dom_element.dispatchEvent(new CustomEvent("update", { detail: null }));
      }
    };

    let options = {
      timeout: 10000,
      enableHighAccuracy: false,
      maximumAge: 60000,
    };

    map.gpsManager.getCurrentPosition(position => {
      position_callback(position);
      // Try to upgrade to a high accuracy position
      options.enableHighAccuracy = true;
      options.maximumAge = 500;
      this.geolocation_watch_id = map.gpsManager.watchPosition(position_callback.bind(this), error_callback, options);
    }, error_callback, options);
  }

  async #lookupWaypoint(osm_id) {
    let params = new URLSearchParams({
      osm_ids: osm_id,
      format: "json",
      dedupe: 1,
      extratags: 1,
      addressdetails: 1,
      namedetails: 1,
      polygon_geojson: 1,
    });
    let base = "/lookup";
    return fetch(base + "?" + params.toString())
      .then((result) => result.json())
      .then((result) => {
        if (!result || result.length == 0) {
          throw new Error(`Cannot find Result for OSM Object: "${osm_id}"`);
        }
        this.data = result[0];
        return;
      });
  }

  #createDomElement() {
    // Outer result container
    if (!this.dom_element) {
      this.dom_element = document.createElement("div");
      this.dom_element.classList.add("result");
    } else {
      this.dom_element.textContent = "";
    }

    // Create loading content if data is not loaded yet
    if (!this.data && !this.error_message) {
      let img = document.createElement("img");
      img.src = "/img/ajax-loader.gif";
      this.dom_element.appendChild(img);
      let text = document.createElement("div");
      text.textContent = "Lade Wegpunkt...";
      this.dom_element.appendChild(text);
      return;
    } else {
      this.dom_element.id = `result-${this.rank}`;
    }

    // Reorder waypoints
    let reorder_container = document.createElement("div");
    reorder_container.classList.add("reorder-waypoints");
    let up_container = document.createElement("a");
    up_container.classList.add("up");
    up_container.title = "Wegpunkt nach oben verschieben";
    up_container.innerHTML = "&#x25B2;";
    reorder_container.appendChild(up_container);
    let down_container = document.createElement("a");
    down_container.classList.add("down");
    down_container.title = "Wegpunkt nach unten verschieben";
    down_container.innerHTML = "&#x25BC;";
    reorder_container.appendChild(down_container);
    this.dom_element.appendChild(reorder_container);

    // Inner Marker
    let marker_container = document.createElement("div");
    marker_container.classList.add("result-marker");
    this.dom_element.appendChild(marker_container);

    let removeWaypoint = document.createElement("a");
    removeWaypoint.classList.add("remove-waypoint");
    removeWaypoint.title = "Wegpunkt löschen";
    removeWaypoint.innerHTML = "&#128465;";
    this.dom_element.appendChild(removeWaypoint);

    if (this.error_message) {
      let warning_container = document.createElement("div");
      warning_container.classList.add("warning-container");
      this.dom_element.appendChild(warning_container);
      let warning_sign = document.createElement("div");
      warning_sign.innerHTML = "&#9888;";
      warning_sign.classList.add("warning-sign");
      warning_container.appendChild(warning_sign);
      let message = document.createElement("div");
      message.classList.add("message");
      message.textContent = this.error_message;
      warning_container.appendChild(message);
    } else {
      // Inner display name
      let display_name_container = document.createElement("div");
      display_name_container.classList.add("display-name");
      display_name_container.textContent = this.data.display_name;
      this.dom_element.appendChild(display_name_container);
    }
    let addWaypointButton = document.createElement("a");
    if (!this.actions_enabled) {
      addWaypointButton.classList.add("hidden");
    }
    addWaypointButton.classList.add("button", "add-waypoint");
    // Todo switch text to "add waypoint when route is already initialized"
    if (map.routeModule.waypoints.length > 0) {
      addWaypointButton.textContent = "+ Wegpunkt hinzufügen";
    } else {
      addWaypointButton.textContent = "Route berechnen";
    }
    this.dom_element.appendChild(addWaypointButton);
  }

  #createMarker() {
    if (!this.data && !this.error_message) return;
    if (!this.data.lon || !this.data.lat) return;
    let chr = String.fromCharCode(65 + this.rank);
    let huerotate = Math.round(this.rank * 12.5) + 30;
    let el = document.createElement("span");
    el.id = `result-marker-map-${chr}`;
    el.classList.add("marker");
    el.textContent = chr;
    if (huerotate) {
      el.style.filter = `hue-rotate(${huerotate}deg)`;
    }

    let el_result = el.cloneNode(true);
    el_result.id = `result-marker-result-${chr}`;

    if (this.marker) {
      this.marker.remove();
    }
    this.marker = new maplibregl.Marker({
      element: el,
    });
    if (this.data) {
      this.marker.setLngLat([this.data.lon, this.data.lat]);
    }

    this.dom_element.querySelector(".result-marker").appendChild(el_result);

    this.dom_element.addEventListener("mouseover", (e) => {
      el.classList.add("hover");
      this.dom_element.classList.add("hover");
      el.style.filter = "";
      el_result.style.filter = "";
    });
    this.dom_element.addEventListener("mouseout", (e) => {
      el.classList.remove("hover");
      this.dom_element.classList.remove("hover");
      el.style.filter = `hue-rotate(${huerotate}deg)`;
      el_result.style.filter = `hue-rotate(${huerotate}deg)`;
    });
    el.addEventListener("mouseover", (e) => {
      el.classList.add("hover");
      this.dom_element.classList.add("hover");
      el_result.classList.add("hover");
      el_result.style.filter = "";
    });
    el.addEventListener("mouseout", (e) => {
      el.classList.remove("hover");
      this.dom_element.classList.remove("hover");
      el_result.classList.remove("hover");
      el.style.filter = `hue-rotate(${huerotate}deg)`;
      el_result.style.filter = `hue-rotate(${huerotate}deg)`;
    });

    return el;
  }
}
