class RouteFinder {
  htmlmodule = document.getElementById("route-finder-addon");
  searchinput = this.htmlmodule.querySelector("input#search-waypoint-input");
  navigation_active = false;
  // Callbacks
  moveend_callback = this.updateState.bind(this);
  click_callback = this.handleMapClick.bind(this);

  lastQuery = "";

  vehicle = "guess";
  waypoints = [];
  routes = [];

  resultHistory = new LocalHistory("results");
  searchHistory = new LocalHistory("suche");

  mapsourcename = "routefinder"; // Source where all Features of the Routefinder are displayed in
  linesourcename = `${this.mapsourcename}-line`;
  fillsourcename = `${this.mapsourcename}-fill`;
  inactivesourcename = `${this.mapsourcename}-inactive-line`;
  markers = [];

  constructor() {
    this.addMapLayers();

    this.searchinput.onkeyup = (e) => {
      let query = this.htmlmodule.querySelector(
        "input#search-waypoint-input"
      ).value;
      map.switchModule("search", {
        query: query,
        search: true,
        stateupdates: false,
        exit_callback: () => {
          map.switchModule("route-finding");
        },
      });
    };

    this.htmlmodule.querySelector(".total-info .focus").onclick =
      this.fitRoutesOnMap.bind(this);

    this.htmlmodule
      .querySelectorAll("input[name=vehicle]")
      .forEach((element) => {
        element.onchange = (event) => {
          this.vehicle = element.value;
          this.calculateRoutes()
            .then(() => this.updateInterface())
            .then(() => this.fitRoutesOnMap());
        };
      });

    this.htmlmodule.querySelector("#start-navigation").onclick = (e) =>
      this.startNavigation();

    // Add EventListener to Exit Button
    this.htmlmodule.querySelector("#vehicle-chooser >button.close").onclick = (
      e
    ) => {
      // Remove all waypoints
      while (this.waypoints.length > 0) {
        this.removeWaypoint(0);
      }
      map.switchModule("search");
    };
  }

  async enable(
    vehicle = null,
    waypoints = [],
    mapposition = null,
    append = true,
    navigation_active = false
  ) {
    this.htmlmodule.classList.add("active");
    this.searchinput.value = "";

    map.map.off("moveend", this.moveend_callback);
    map.map.on("moveend", this.moveend_callback);

    map.map.off("click", this.click_callback);
    map.map.on("click", this.click_callback);

    if (vehicle) this.vehicle = vehicle;
    this.navigation_active = navigation_active;
    // supplied waypoints will be added to the existing list unless append is defined and false
    if (!append) {
      while (this.waypoints.length > 0) {
        let waypoint = this.waypoints.pop();
        waypoint.removeMarker();
      }
    }
    // Add GPS options
    await map.gpsManager.isGPSAvailable().then((gps_available) => {
      this.htmlmodule.classList.remove("navigation-enabled");
      if (
        gps_available &&
        waypoints.length + this.waypoints.length == 1 &&
        waypoints[0] !== "gps"
      ) {
        waypoints.unshift("gps");
      }
    });
    for (const [index, waypoint] of waypoints.entries()) {
      if (waypoint instanceof Result) {
        this.addWaypoint(waypoint);
        await this.calculateRoutes();
        this.updateInterface();
        continue;
      }
      let detail_match = waypoint.match(/^(N|W|R)(\d+)$/);
      let new_waypoint = null;
      if (detail_match) {
        new_waypoint = new Result(
          detail_match[1] + detail_match[2],
          false,
          "lookup"
        );
      } else if (waypoint.startsWith("gps")) {
        new_waypoint = new Result(waypoint, false, "gps");
      }
      if (new_waypoint) {
        new_waypoint.rank = this.waypoints.length;
        new_waypoint.dom_element.addEventListener("load", async (e) => {
          let all_routes_calculated = await this.calculateRoutes();
          this.updateInterface();
          if (!mapposition) {
            this.fitRoutesOnMap();
          }
          if (this.navigation_active && all_routes_calculated) {
            this.startNavigation();
          }
        }, { once: true });
        this.addWaypoint(new_waypoint);
      }
    }

    if (waypoints.length == 0) {
      this.updateInterface();
      this.fitRoutesOnMap();
    }

    if (mapposition) {
      mapposition = new maplibregl.LngLatBounds(
        mapposition._sw,
        mapposition._ne
      );
      let pos = map.map.cameraForBounds(mapposition);
      map.map.easeTo(
        {
          center: pos.center,
          zoom: pos.zoom,
        },
        { user: false }
      );
    } else {
      this.fitRoutesOnMap();
    }
  }

  removeWaypoint(index) {
    if (this.waypoints.length < index - 1) return;
    this.waypoints[index].removeMarker();

    if (this.waypoints[index].geolocation_watch_id) {
      map.gpsManager.clearWatch(this.waypoints[index].geolocation_watch_id);
    }
    if (this.waypoints[index].route) this.waypoints[index].route.hidePopups();
    if (index > 0) {
      this.waypoints[index - 1].route.hidePopups();
      this.waypoints[index - 1].route = null;
    }
    this.waypoints.splice(index, 1);
    this.updateWaypointRanks();
  }

  addWaypoint(waypoint, first = false) {
    waypoint.dom_element.classList.add("route-finder");
    if (first) {
      this.waypoints.unshift(waypoint);
    } else {
      this.waypoints.push(waypoint);
    }
    this.updateWaypointRanks();
    this.updateInterface();
    // Register Event handler for gps updates
    if (waypoint.rank == 0 && waypoint.source == "gps") {
      waypoint.dom_element.addEventListener("update", this.gpsUpdate.bind(this));
      waypoint.dom_element.addEventListener("navigation_available", e => {
        this.htmlmodule.classList.add("navigation-enabled");
      });
      // Check if Navigation already is set to available
      if (waypoint.navigation_available) {
        waypoint.dom_element.dispatchEvent(new Event("navigation_available"));
      }
    }
  }

  gpsUpdate(e) {
    this.updateInterface();
    // Check if the route needs to be recalculated
  }

  // Go through all waypoints and validate that the stored rank is correct
  updateWaypointRanks() {
    for (const [index, waypoint] of this.waypoints.entries()) {
      if (waypoint.rank != index) {
        waypoint.updateRank(index);
      }
    }
  }

  async calculateRoutes() {
    let routes_calculated = 0;
    for (const [index, waypoint] of this.waypoints.entries()) {
      // If waypoint is the last make sure there is no route calculated
      if (index == this.waypoints.length - 1) {
        delete waypoint.route;
        continue;
      }
      // Check if the waypoint has a route calculated and is correct
      if (
        waypoint.route &&
        waypoint.route.profile == this.vehicle &&
        waypoint.route.waypoint_from_id == waypoint.toString() &&
        waypoint.route.waypoint_to_id == this.waypoints[index + 1].toString()
      ) {
        routes_calculated++;
        continue;
      }

      // Check if both waypoints are already loaded
      if (!waypoint.data || !this.waypoints[index + 1].data) {
        delete waypoint.route;
        continue;
      }
      if (waypoint.route) waypoint.route.hidePopups();
      waypoint.route = new Route(this.vehicle);

      await waypoint.route
        .calculateRoute(this.waypoints[index], this.waypoints[index + 1])
        .then(() => {
          if (this.vehicle == "guess") {
            this.vehicle = waypoint.route.profile;
            this.htmlmodule.querySelector(
              `input[type=radio][value=${this.vehicle}]`
            ).checked = true;
          }
        });
      routes_calculated++;
    }
    if (routes_calculated == this.waypoints.length - 1) {
      return true;
    } else {
      return false;
    }
  }

  updateInterface() {
    this.clearInterface();
    this.updateRouteContainer();
    this.updateMap();
  }

  updateRouteContainer() {
    if (this.waypoints.length > 2) {
      // Update total route information
      let route_count = 0;
      let duration = 0;
      let distance = 0;
      for (const waypoint of this.waypoints) {
        if (!waypoint.route || !waypoint.route.data) continue;
        route_count++;
        duration += waypoint.route.data.paths[waypoint.route.selected].time;
        distance += waypoint.route.data.paths[waypoint.route.selected].distance;
      }
      this.htmlmodule.querySelector(".total-info .duration").textContent =
        formatDuration(duration);
      this.htmlmodule.querySelector(".total-info .distance").textContent =
        formatDistance(distance);
      if (route_count >= 2) this.htmlmodule.classList.add("multiroute");
    } else {
      this.htmlmodule.classList.remove("multiroute");
    }

    // Update vehicle
    if (this.vehicle != "guess")
      this.htmlmodule.querySelector(
        `input[name=vehicle][value=${this.vehicle}`
      ).checked = true;

    let route_interface_container = this.htmlmodule.querySelector(
      "#waypoint-list-container"
    );

    // Add a button to add GPS Position if available
    let gps_added = false;
    for (const waypoint of this.waypoints) {
      if (waypoint.source == "gps") gps_added = true;
    }
    if (!gps_added) {
      let add_gps_position_container = document.createElement("a");
      add_gps_position_container.classList.add("add-gps");
      add_gps_position_container.textContent = "+ Eigene Position hinzufügen";
      route_interface_container.appendChild(add_gps_position_container);
      add_gps_position_container.onclick = (e) => {
        let gps_waypoint = new Result(
          "gps",
          false,
          "gps",
          0,
          null,
          async () => {
            setTimeout(async () => {
              await this.calculateRoutes();
              this.updateInterface();
              this.updateState();
            }, 250);
          }
        );
        this.addWaypoint(gps_waypoint, true);
      };
    }

    for (const [index, waypoint] of this.waypoints.entries()) {
      if (waypoint.error_code == Result.ERROR_GEOLOCATION_PERMISSION_DENIED || waypoint.error_code == Result.ERROR_GEOLOCATION_TIMEOUT) {
        // User denied Geolocation Access. Remove GPS waypoint
        waypoint.removeMarker();
        this.waypoints.splice(index, 1);
        this.updateInterface();
        this.updateState();
        return;
      }
      route_interface_container.appendChild(waypoint.dom_element);
      if (waypoint.data) {
        // Delete Waypoint callback
        waypoint.dom_element.querySelector(".remove-waypoint").onclick = (
          e
        ) => {
          this.removeWaypoint(index);
          this.updateWaypointRanks();
          this.calculateRoutes()
            .then(() => this.updateInterface())
            .then(() => this.fitRoutesOnMap())
            .then(() => this.updateState());
        };
        waypoint.dom_element
          .querySelectorAll(".reorder-waypoints > a")
          .forEach((element) => {
            element.onclick = (click) => {
              let up = click.target.classList.contains("down") ? false : true;
              if (up && index == 0) return; // Do not swap up if first element
              if (!up && index == this.waypoints.length - 1) return; // Do not swap down if last element
              // Swap waypoint indexes
              if (up) {
                [this.waypoints[index - 1], this.waypoints[index]] = [
                  this.waypoints[index],
                  this.waypoints[index - 1],
                ];
              } else {
                [this.waypoints[index], this.waypoints[index + 1]] = [
                  this.waypoints[index + 1],
                  this.waypoints[index],
                ];
              }
              this.updateWaypointRanks();
              this.calculateRoutes()
                .then(() => this.updateInterface())
                .then(() => this.fitRoutesOnMap())
                .then(() => this.updateState());
            };
          });
      }
      if (waypoint.route && waypoint.route.data) {
        route_interface_container.appendChild(waypoint.route.dom_node);
        waypoint.route.dom_node.onchange = () => {
          this.updateInterface();
        };
      }
    }
  }
  clearRouteContainer() {
    let route_interface_container = this.htmlmodule.querySelector(
      "#waypoint-list-container"
    );
    route_interface_container.textContent = "";
  }

  /**
   * Prints all waypoints and routes on the map
   */
  updateMap() {
    // Add the markers to the map
    for (const [index, waypoint] of this.waypoints.entries()) {
      if (!this.navigation_active || index > 0) waypoint.showMarker();
      waypoint.showGeoJson(this.linesourcename, this.fillsourcename);

      if (waypoint.route) {
        waypoint.route.showPopups();
        waypoint.route.showGeoJson(
          this.linesourcename,
          this.inactivesourcename
        );
      }
    }
  }
  clearMap() {
    for (const waypoint of this.waypoints) {
      waypoint.removeMarker();
      if (waypoint.route) waypoint.route.hidePopups();
    }
    let line_data = map.map.getSource(this.linesourcename)._data;
    line_data.features = [];
    map.map.getSource(this.linesourcename).setData(line_data);

    let inactive_line_data = map.map.getSource(this.inactivesourcename)._data;
    inactive_line_data.features = [];
    map.map.getSource(this.inactivesourcename).setData(inactive_line_data);

    let fill_data = map.map.getSource(this.fillsourcename)._data;
    fill_data.features = [];
    map.map.getSource(this.fillsourcename).setData(fill_data);
  }

  startNavigation() {
    if (this.waypoints[0].source != "gps" || !this.waypoints[0].route) return;
    map.gpsManager.clearWatch(this.waypoints[0].geolocation_watch_id);
    this.waypoints[0].geolocation_watch_id = null;

    this.navigation_active = true;
    this.htmlmodule.classList.remove("active"); // Route Overview is not shown while navigating
    map.map.off("click", this.click_callback); // No reverse searches whilst navigating
    map.map.off("moveend", this.moveend_callback); // No state updates whilst navigating
    this.updateState(); // But a final state update
    map.navigationModule.enable();
    this.updateInterface();
  }
  stopNavigation() {
    this.navigation_active = false;
    this.htmlmodule.classList.add("active"); // Route Overview is not shown while navigating
    map.map.on("click", this.click_callback); // No reverse searches whilst navigating
    map.map.on("moveend", this.moveend_callback); // No state updates whilst navigating
    this.updateState(); // But a final state update
  }
  handleMapClick(event) {
    let search_params = {
      reverse: {
        lngLat: event.lngLat,
        zoom: Math.round(map.map.getZoom()),
      },
      stateupdates: false,
      search: true,
      exit_callback: () => {
        map.switchModule("route-finding");
      },
    };
    map.switchModule("search", search_params);
  }

  fitRoutesOnMap() {
    if (this.waypoints.length <= 1) return;
    let bbox_collection = null;
    for (const waypoint of this.waypoints) {
      if (!waypoint.route) continue;
      let bbox = waypoint.route.data.paths[waypoint.route.selected].bbox;
      bbox = turf.bboxPolygon([bbox[0], bbox[1], bbox[2], bbox[3]]);
      if (bbox_collection)
        bbox_collection = turf.featureCollection([...bbox_collection, bbox]);
      else bbox_collection = turf.featureCollection([bbox]);
    }
    let bbox = turf.bbox(bbox_collection);
    map.fitBbox([
      [bbox[0], bbox[1]],
      [bbox[2], bbox[3]],
    ]);
  }

  addMapLayers() {
    // Variables for map features
    let lineSource = {
      type: "geojson",
      data: {
        type: "FeatureCollection",
        features: [],
      },
    };
    let lineLayer = {
      id: `${this.linesourcename}`,
      type: "line",
      source: `${this.linesourcename}`,
      paint: {
        "line-color": "#2e8bc0",
        "line-width": 5,
        "line-opacity": 1,
      },
    };

    let inactiveLineSource = {
      type: "geojson",
      data: {
        type: "FeatureCollection",
        features: [],
      },
    };
    let inactiveLineLayer = {
      id: `${this.inactivesourcename}`,
      type: "line",
      source: `${this.inactivesourcename}`,
      paint: {
        "line-color": "rgb(177, 180, 185)",
        "line-width": 5,
        "line-opacity": 0.6,
      },
    };

    let fillSource = {
      type: "geojson",
      data: {
        type: "FeatureCollection",
        features: [],
      },
    };
    let fillLayer = {
      id: `${this.fillsourcename}`,
      type: "fill",
      source: `${this.fillsourcename}`,
      paint: {
        "fill-color": "rgb(255,127,0)",
        "fill-outline-color": "rgb(255,127,0)",
        "fill-opacity": 0.3,
      },
    };

    let addLayers = () => {
      map.map.addSource(this.inactivesourcename, inactiveLineSource);
      map.map.addLayer(inactiveLineLayer);
      map.map.addSource(this.linesourcename, lineSource);
      map.map.addLayer(lineLayer);

      map.map.addSource(this.fillsourcename, fillSource);
      map.map.addLayer(fillLayer);
    };
    if (map.map.isStyleLoaded()) {
      addLayers();
    } else {
      map.map.on("load", addLayers);
    }
  }

  clearInterface() {
    this.clearRouteContainer();
    this.clearMap();
  }
  /**
   * Function called when map is loaded and GPS Features become available
   */
  enableGps() { }

  /**
   * Cleanup Code when module is closed
   */
  exit() {
    this.htmlmodule.classList.remove("active");

    this.clearInterface();

    map.map.off("moveend", this.moveend_callback);
    map.map.off("click", this.click_callback);
  }

  updateState(event) {
    if (
      event &&
      ((event.hasOwnProperty("user") && event.user == false) ||
        (event.hasOwnProperty("0") &&
          ResizeObserverEntry.prototype.isPrototypeOf(event[0])))
    ) {
      return;
    }
    let url_vehicle = this.vehicle !== undefined ? this.vehicle : "guess";
    let url = window.location.origin + `/route/${url_vehicle}/`;

    let waypoint_strings = [];
    for (const waypoint of this.waypoints) {
      waypoint_strings.push(waypoint.toString());
    }
    url += waypoint_strings.join(";");

    let current_pos = map.map.getBounds();

    // Check if current_pos changed enough
    if (
      window.history &&
      window.history.state &&
      window.history.state.module == "route-finder" &&
      window.history.state.navigation_active == this.navigation_active
    ) {
      let old_pos = new maplibregl.LngLatBounds(
        window.history.state.mapposition._sw,
        window.history.state.mapposition._ne
      );
      let old_pos_camera = map.map.cameraForBounds(old_pos);
      let new_pos_camera = map.map.cameraForBounds(current_pos);
      if (
        old_pos_camera.center.distanceTo(new_pos_camera.center) <= 250 &&
        Math.abs(old_pos_camera.zoom - new_pos_camera.zoom) < 0.5
      ) {
        current_pos = old_pos;
      }
    }

    url += `/${JSON.stringify(current_pos.toArray())}`;

    if (this.navigation_active) {
      url += "?navigation";
    }

    // Make Waypoints serializable i.e. strip any dom nodes
    let waypoints = [];
    for (const waypoint of this.waypoints) {
      let route = null;
      if (waypoint.route) {
        route = {
          data: waypoint.route.data,
          profile: waypoint.route.profile,
        };
      }
      waypoints.push({
        data: waypoint.data,
        route: route,
        source: waypoint.source,
      });
    }

    let state = {
      module: "route-finder",
      mapposition: current_pos,
      waypoints: waypoints,
      vehicle: this.vehicle,
      navigation_active: this.navigation_active,
      timestamp: Date.now(),
    };

    if (
      window.history.state == undefined ||
      window.location.href == url ||
      Date.now() - window.history.state.timestamp < 100
    ) {
      window.history.replaceState(state, "", url);
    } else {
      window.history.pushState(state, "", url);
    }
  }
}
